import unittest
from sismic import io
from sismic.interpreter import Interpreter
from sismic.model import MacroStep, MicroStep
from sismic.stories import *


class StoryTests(unittest.TestCase):
    def test_storycreation(self):
        story = Story()
        story.append(Event('a'))
        story.append(Event('b'))
        story.append(Pause(10))
        story.append(Event('c'))

        self.assertEqual(len(story), 4)

    def test_tell(self):
        story = Story([Event('goto s2'), Pause(5), Event('goto final')])
        sc = io.import_from_yaml(open('tests/yaml/simple.yaml'))
        interpreter = Interpreter(sc)
        story.tell(interpreter)

        self.assertTrue(interpreter.final)
        self.assertEqual(interpreter.time, 5)
        self.assertEqual(len(interpreter.trace), 4)


class RandomStoryTests(unittest.TestCase):
    def setUp(self):
        self.story = Story([Event('a'), Event('b'), Event('c'), Pause(1), Pause(2)])

    def test_length(self):
        self.assertEqual(len(next(random_stories_generator(self.story))), 5)
        self.assertEqual(len(next(random_stories_generator(self.story, length=3))), 3)
        self.assertEqual(len(next(random_stories_generator(self.story, length=6))), 6)
        self.assertEqual(len(next(random_stories_generator([]))), 0)

    def test_number(self):
        self.assertEqual(len(list(random_stories_generator(self.story, number=10))), 10)


class StoryFromTraceTests(unittest.TestCase):
    def test_empty_trace(self):
        trace = []
        self.assertListEqual(story_from_trace(trace), [])

    def test_events(self):
        trace = [
            MacroStep(0, [MicroStep(Event('a'))]),
            MacroStep(0, [MicroStep(Event('b'))]),
            MacroStep(0, [MicroStep(Event('c'))]),
            MacroStep(0, [MicroStep(Event('d'))]),
        ]
        self.assertListEqual(story_from_trace(trace), [
            Event('a'), Event('b'), Event('c'), Event('d')
        ])

    def test_pauses(self):
        trace = [
            MacroStep(0, []),
            MacroStep(10, []),
            MacroStep(10, []),
            MacroStep(15, []),
        ]
        self.assertListEqual(story_from_trace(trace), [
            Pause(10), Pause(5)
        ])

    def test_initial_pause(self):
        trace = [MacroStep(10, [])]
        self.assertListEqual(story_from_trace(trace), [Pause(10)])

    def test_events_and_pauses(self):
        trace = [
            MacroStep(2, [MicroStep(Event('a'))]),
            MacroStep(5, [MicroStep(Event('b'))]),
            MacroStep(9, [MicroStep(Event('c'))]),
            MacroStep(14, [MicroStep(Event('d'))]),
        ]
        self.assertListEqual(story_from_trace(trace), [
            Pause(2), Event('a'), Pause(3), Event('b'), Pause(4), Event('c'), Pause(5), Event('d')
        ])



import abc
import copy
from .model import Event, Transition, StateMixin


class Evaluator(metaclass=abc.ABCMeta):
    """
    Abstract base class for any evaluator.

    An instance of this class defines what can be done with piece of codes
    contained in a statechart (condition, action, etc.).

    Notice that the execute_* methods are called at each step, even if there is no
    code to execute. This allows the evaluator to keep track of the states that are
    entered or exited, and of the transitions that are processed.

    :param interpreter: the interpreter that will use this evaluator,
        is expected to be an ``Interpreter`` instance
    :param initial_context: an optional dictionary to populate the context
    """
    def __init__(self, interpreter=None, initial_context: dict = None):
        self._context = initial_context if initial_context else {}
        self._interpreter = interpreter

    @property
    def context(self) -> dict:
        """
        The context of this evaluator. A context is a dict-like mapping between
        variables and values that is expected to be exposed when the code is evaluated.
        """
        return self._context

    @abc.abstractmethod
    def _evaluate_code(self, code: str, additional_context: dict = None) -> bool:
        """
        Generic method to evaluate a piece of code. This method is a fallback if one of
        the other evaluate_* methods is not overridden.

        :param code: code to evaluate
        :param additional_context: an optional additional context
        :return: truth value of *code*
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _execute_code(self, code: str, additional_context: dict = None):
        """
        Generic method to execute a piece of code. This method is a fallback if one
        of the other execute_* methods is not overridden.

        :param code: code to execute
        :param additional_context: an optional additional context
        """
        raise NotImplementedError()

    def evaluate_guard(self, transition: Transition, event: Event) -> bool:
        """
        Evaluate the guard for given transition.

        :param transition: the considered transition
        :param event: instance of ``Event`` if any
        :return: truth value of *code*
        """
        if transition.guard:
            return self._evaluate_code(transition.guard, {'event': event})

    def execute_action(self, transition: Transition, event: Event) -> bool:
        """
        Execute the action for given transition.
        This method is called for every transition that is processed, even those with no ``action``.

        :param transition: the considered transition
        :param event: instance of ``Event`` if any
        :return: truth value of *code*
        """
        if transition.action:
            return self._execute_code(transition.action, {'event': event})

    def execute_onentry(self, state: StateMixin):
        """
        Execute the on entry action for given state.
        This method is called for every state that is entered, even those with no ``on_entry``.

        :param state: the considered state
        """
        if getattr(state, 'on_entry', None):
            return self._execute_code(state.on_entry)

    def execute_onexit(self, state: StateMixin):
        """
        Execute the on exit action for given state.
        This method is called for every state that is exited, even those with no ``on_exit``.

        :param state: the considered state
        """
        if getattr(state, 'on_exit', None):
            return self._execute_code(state.on_exit)

    def evaluate_preconditions(self, obj, event: Event = None) -> list:
        """
        Evaluate the preconditions for given object (either a ``StateMixin`` or a
        ``Transition``) and return a list of conditions that are not satisfied.
        
        :param obj: the considered state or transition
        :param event: an optional ``Event`` instance, in the case of a transition
        :return: list of unsatisfied conditions
        """
        event_d = {'event': event} if isinstance(obj, Transition) else None
        return filter(lambda c: not self._evaluate_code(c, event_d), getattr(obj, 'preconditions', []))

    def evaluate_invariants(self, obj, event: Event = None) -> list:
        """
        Evaluate the invariants for given object (either a ``StateMixin`` or a
        ``Transition``) and return a list of conditions that are not satisfied.
        
        :param obj: the considered state or transition
        :param event: an optional ``Event`` instance, in the case of a transition
        :return: list of unsatisfied conditions
        """
        event_d = {'event': event} if isinstance(obj, Transition) else None
        return filter(lambda c: not self._evaluate_code(c, event_d), getattr(obj, 'invariants', []))

    def evaluate_postconditions(self, obj, event: Event = None) -> list:
        """
        Evaluate the postconditions for given object (either a ``StateMixin`` or a
        ``Transition``) and return a list of conditions that are not satisfied.
        
        :param obj: the considered state or transition
        :param event: an optional ``Event`` instance, in the case of a transition
        :return: list of unsatisfied conditions
        """
        event_d = {'event': event} if isinstance(obj, Transition) else None
        return filter(lambda c: not self._evaluate_code(c, event_d), getattr(obj, 'postconditions', []))


class DummyEvaluator(Evaluator):
    """
    A dummy evaluator that does nothing and evaluates every condition to True.
    """

    @property
    def context(self):
        return dict()

    def _evaluate_code(self, code: str, additional_context: dict = None) -> bool:
        return True

    def _execute_code(self, code: str, additional_context: dict = None):
        return


class PythonEvaluator(Evaluator):
    """
    Evaluator that interprets Python code.

    Depending on the method that is called, the context can expose additional values:

    - Always:
        - A ``time`` value that represents the current time exposed by the interpreter.
        - An ``active(name) -> bool`` Boolean function that takes a state name and return ``True`` if and only
          if this state is currently active, ie. it is in the active configuration of the ``Interpreter`` instance
          that makes use of this evaluator.
    - On code execution:
        - A ``send`` function that takes an event name and additional keyword parameters and fires an internal event with.
        - If the code is related to a transition, the ``event`` that fires the transition is exposed.
    - On code evaluation:
        - If the code is related to a transition, the ``event`` that fires the transition is exposed.
    - On guard evaluation:
        - An ``after(sec) -> bool`` Boolean function that returns ``True`` if and only if the source state
          was entered more than *sec* seconds ago. The time is evaluated according to Interpreter's clock.
        - A ``idle(sec) -> bool`` Boolean function that returns ``True`` if and only if the source state
          did not fire a transition for more than *sec* ago. The time is evaluated according to Interpreter's clock.
    - On postcondition or invariant:
        - A variable ``__old__`` that has an attribute ``x`` for every ``x`` in the context when either the state
          was entered (if the condition involves a state) or the transition was processed (if the condition
          involves a transition). The value of ``__old__.x`` is a shallow copy of ``x`` at that time.

    Unless you override its entry in the context, the ``__builtins__`` of Python are automatically exposed.
    This implies you can use nearly everything from Python in your code.

    If an exception occurred while executing or evaluating a piece of code, it is propagated by the
    evaluator.

    :param interpreter: the interpreter that will use this evaluator,
        is expected to be an ``Interpreter`` instance
    :param initial_context: a dictionary that will be used as ``__locals__``
    """

    def __init__(self, interpreter=None, initial_context: dict = None):
        super().__init__(interpreter, initial_context)
        self._memory = {}  # Associate to each state or transition the context on state entry and transition action
        self._idle_time = {}  # Associate a timer to each state name (idle timer)
        self._entry_time = {}  # Associate a timer to each state name (entry timer)

    def _evaluate_code(self, code: str, additional_context: dict = None) -> bool:
        """
        Evaluate given code using Python.

        :param code: code to evaluate
        :param additional_context: an optional additional context
        :return: truth value of *code*
        """
        if not code:
            return True

        exposed_context = {
            'active': lambda s: s in self._interpreter.configuration,
            'time': getattr(self._interpreter, 'time', 0)
        }
        exposed_context.update(additional_context if additional_context else {})

        try:
            return eval(code, exposed_context, self._context)
        except Exception as e:
            raise type(e)('The above exception occurred while evaluating:\n{}'.format(code)) from e

    def _execute_code(self, code: str, additional_context: dict = None):
        """
        Execute given code using Python.

        :param code: code to execute
        :param additional_context: an optional additional context
        """
        if not code:
            return

        def send(name, **kwargs): self._interpreter.send(Event(name, **kwargs), internal=True)

        exposed_context = {
            'active': lambda s: s in self._interpreter.configuration,
            'time': getattr(self._interpreter, 'time', 0),
            'send': send,
        }
        exposed_context.update(additional_context if additional_context else {})

        try:
            exec(code, exposed_context, self._context)
        except Exception as e:
            raise type(e)('The above exception occurred while executing:\n{}'.format(code)) from e

    class FrozenContext:
        """
        A shallow copy of a context. The keys of the underlying context are
        exposed as attributes.
        """

        def __init__(self, context: dict):
            self.__frozencontext = {k: copy.copy(v) for k, v in context.items()}

        def __getattr__(self, item):
            return self.__frozencontext[item]

    def evaluate_guard(self, transition: Transition, event: Event) -> bool:
        if transition.guard:
            context = {
                'event': event,
                'after': lambda i: self._interpreter.time - i >= self._entry_time[transition.from_state],
                'idle': lambda i: self._interpreter.time - i >= self._idle_time[transition.from_state]
            }
            return self._evaluate_code(transition.guard, context)

    def execute_onentry(self, state: StateMixin):
        # Set memory
        self._memory[id(state)] = PythonEvaluator.FrozenContext(self._context)

        # Set timer
        self._entry_time[state.name] = self._interpreter.time
        self._idle_time[state.name] = self._interpreter.time

        super().execute_onentry(state)

    def execute_action(self, transition: Transition, event: Event):
        # Set memory
        self._memory[id(transition)] = PythonEvaluator.FrozenContext(self._context)
        # Set timer
        self._idle_time[transition.from_state] = self._interpreter.time

        super().execute_action(transition, event)

    def evaluate_postconditions(self, obj, event: Event = None):
        context = {'event': event} if isinstance(obj, Transition) else {}
        context['__old__'] = (self._memory.get(id(obj), None))

        return filter(lambda c: not self._evaluate_code(c, context), getattr(obj, 'postconditions', []))

    def evaluate_invariants(self, obj, event: Event = None):
        context = {'event': event} if isinstance(obj, Transition) else {}
        context['__old__'] = (self._memory.get(id(obj), None))

        return filter(lambda c: not self._evaluate_code(c, context), getattr(obj, 'invariants', []))



import yaml
import os
from pykwalify.core import Core
from collections import OrderedDict
from .model import Event, Transition, StateChart, BasicState, CompoundState, OrthogonalState, HistoryState, FinalState
from .model import StateMixin, ActionStateMixin, TransitionStateMixin, CompositeStateMixin


SCHEMA_PATH = os.path.join(os.path.dirname(__file__), 'schema.yaml')


def import_from_yaml(statechart: str, validate_yaml=True, validate_statechart=True) -> StateChart:
    """
    Import a statechart from a YAML representation.
    YAML is first validated against ``io.SCHEMA``.

    :param statechart: string or any equivalent object
    :param validate_yaml: set to ``False`` to disable yaml validation.
    :param validate_statechart: set to ``False`` to disable statechart validation
      (see ``model.StateChart.validate``).
    :return: a ``StateChart`` instance
    """
    statechart_data = yaml.load(statechart)
    if validate_yaml:
        checker = Core(source_data=statechart_data, schema_files=[SCHEMA_PATH])
        checker.validate(raise_exception=True)

    sc = _import_from_dict(statechart_data['statechart'])
    if validate_statechart:
        sc.validate()

    return sc


def _import_from_dict(data: dict) -> StateChart:
    """
    Import a statechart from a (set of nested) dictionary.

    :param data: dict-like structure
    :return: a StateChart instance
    """
    sc = StateChart(data['name'], data['initial'], data.get('on entry', None))

    # Preconditions, postconditions and invariants
    for condition in data.get('contract', []):
        if condition.get('before', None):
            sc.preconditions.append(condition['before'])
        elif condition.get('after', None):
            sc.postconditions.append(condition['after'])
        elif condition.get('always', None):
            sc.invariants.append(condition['always'])

    states_to_add = []  # list of (state, parent) to be added
    for state in data['states']:
        states_to_add.append((state, None))

    # Add states
    while states_to_add:
        state_d, parent_name = states_to_add.pop()

        # Create and register state
        try:
            state = _state_from_dict(state_d)
        except Exception as e:
            raise ValueError('An exception occurred while trying to parse:\n {1}\n\nException:\n{0}'.format(e, state_d))
        sc.register_state(state, parent_name)

        # Register transitions if any
        for transition_d in state_d.get('transitions', []):
            try:
                transition = _transition_from_dict(state.name, transition_d)
            except Exception as e:
                raise ValueError('An exception occurred while trying to parse transitions in '
                                 '{2}:\n {1}\n\nException:\n{0}'.format(e, transition_d, state.name))
            sc.register_transition(transition)

        # Register substates
        for substate in state_d.get('states', state_d.get('parallel states', [])):
            states_to_add.append((substate, state.name))

    return sc


def _transition_from_dict(state_name: str, transition_d: dict) -> Transition:
    """
    Return a Transition instance from given dict.

    :param state_name: name of the state in which the transition is defined
    :param transition_d: a dictionary containing transition data
    :return: an instance of Transition
    """
    event = Event(transition_d['event']) if 'event' in transition_d else None
    transition = Transition(state_name, transition_d.get('target', None), event,
                            transition_d.get('guard', None), transition_d.get('action', None))

    # Preconditions, postconditions and invariants
    for condition in transition_d.get('contract', []):
        if condition.get('before', None):
            transition.preconditions.append(condition['before'])
        elif condition.get('after', None):
            transition.postconditions.append(condition['after'])
        elif condition.get('always', None):
            transition.invariants.append(condition['always'])

    return transition


def _state_from_dict(state_d: dict) -> StateMixin:
    """
    Return the appropriate type of state from given dict.

    :param state_d: a dictionary containing state data
    :return: a specialized instance of State
    """
    # Guess the type of state
    if state_d.get('type', None) == 'final':
        # Final pseudo state
        name = state_d.get('name')
        on_entry = state_d.get('on entry', None)
        on_exit = state_d.get('on exit', None)
        state = FinalState(name, on_entry, on_exit)
    elif state_d.get('type', None) == 'shallow history':
        # Shallow history pseudo state
        state = HistoryState(state_d['name'], state_d.get('initial'), deep=False)
    elif state_d.get('type', None) == 'deep history':
        # Deep history pseudo state
        state = HistoryState(state_d['name'], state_d.get('initial'), deep=True)
    else:
        name = state_d.get('name')
        on_entry = state_d.get('on entry', None)
        on_exit = state_d.get('on exit', None)
        if 'states' in state_d:  # Compound state
            initial = state_d.get('initial', None)
            state = CompoundState(name, initial, on_entry, on_exit)
        elif 'parallel states' in state_d:  # Orthogonal state
            state = OrthogonalState(name, on_entry, on_exit)
        else:
            # Simple state
            state = BasicState(name, on_entry, on_exit)

    # Preconditions, postconditions and invariants
    for condition in state_d.get('contract', []):
        if condition.get('before', None):
            state.preconditions.append(condition['before'])
        elif condition.get('after', None):
            state.postconditions.append(condition['after'])
        elif condition.get('always', None):
            state.invariants.append(condition['always'])
    return state


def export_to_yaml(statechart: StateChart) -> str:
    """
    Export given StateChart instance to YAML

    :param statechart:
    :return: A textual YAML representation
    """
    return yaml.dump(_export_to_dict(statechart, ordered=False),
                     width=1000, default_flow_style=False, default_style='"')


def _export_to_dict(statechart: StateChart, ordered=True) -> dict:
    """
    Export given StateChart instance to a dict.

    :param statechart: a StateChart instance
    :param ordered: set to True to use OrderedDict instead
    :return: a dict that can be used in ``import_from_dict``
    """
    d = OrderedDict() if ordered else dict()
    d['name'] = statechart.name
    d['initial'] = statechart.initial
    d['states'] = statechart.children
    if statechart.on_entry:
        d['on entry'] = statechart.on_entry

    preconditions = getattr(statechart, 'preconditions', [])
    postconditions = getattr(statechart, 'postconditions', [])
    invariants = getattr(statechart, 'invariants', [])
    if preconditions or postconditions or invariants:
        conditions = []
        for condition in preconditions:
            conditions.append({'before': condition})
        for condition in postconditions:
            conditions.append({'after': condition})
        for condition in invariants:
            conditions.append({'always': condition})
        d['contract'] = conditions

    statelist_to_expand = [d['states']]
    while statelist_to_expand:
        statelist = statelist_to_expand.pop()
        for i, state in enumerate(statelist):
            statelist[i] = _export_element_to_dict(statechart.states[state], ordered)
            new_statelist = statelist[i].get('states', statelist[i].get('parallel states', []))
            if len(new_statelist) > 0:
                statelist_to_expand.append(new_statelist)
    return {'statechart': d}


def _export_element_to_dict(el, ordered=False) -> dict:
    """
    Export an element (State, Transition, etc.) to a dict.
    Is used in ``export_to_dict`` to generate a global representation.

    :param el: an instance of ``model.*``
    :param ordered: set to True to use OrderedDict instead of dict
    :return: a dict
    """
    d = OrderedDict() if ordered else dict()

    if isinstance(el, Transition):
        if not el.internal:
            d['target'] = el.to_state
        if not el.eventless:
            d['event'] = el.event.name
        if el.guard:
            d['guard'] = el.guard
        if el.action:
            d['action'] = el.action
    if isinstance(el, StateMixin):
        d['name'] = el.name
    if isinstance(el, CompoundState):
        if el.initial:
            d['initial'] = el.initial
    if isinstance(el, ActionStateMixin):
        if el.on_entry:
            d['on entry'] = el.on_entry
        if el.on_exit:
            d['on exit'] = el.on_exit

    preconditions = getattr(el, 'preconditions', [])
    postconditions = getattr(el, 'postconditions', [])
    invariants = getattr(el, 'invariants', [])
    if preconditions or postconditions or invariants:
        conditions = []
        for condition in preconditions:
            conditions.append({'before': condition})
        for condition in postconditions:
            conditions.append({'after': condition})
        for condition in invariants:
            conditions.append({'always': condition})
        d['contract'] = conditions

    if isinstance(el, TransitionStateMixin):
        if len(el.transitions) > 0:
            d['transitions'] = [_export_element_to_dict(t) for t in el.transitions]
    if isinstance(el, CompositeStateMixin):
        if isinstance(el, OrthogonalState):
            d['parallel states'] = el.children
        else:
            d['states'] = el.children
    if isinstance(el, HistoryState):
        if el.deep:
            d['type'] = 'deep history'
        else:
            d['type'] = 'shallow history'
        if el.initial:
            d['initial'] = el.initial
    if isinstance(el, FinalState):
        d['type'] = 'final'
    return d



import unittest
from sismic import evaluator
from sismic.model import Event


class PythonEvaluatorTests(unittest.TestCase):
    def setUp(self):
        context = {
            'x': 1,
            'y': 2,
            'z': 3
        }
        self.e = evaluator.PythonEvaluator(initial_context=context)

    def test_condition(self):
        self.assertTrue(self.e._evaluate_code('True'))
        self.assertFalse(self.e._evaluate_code('False'))
        self.assertTrue(self.e._evaluate_code('1 == 1'))
        self.assertTrue(self.e._evaluate_code('x == 1'))
        with self.assertRaises(Exception):
            self.e._evaluate_code('a')

    def test_condition_on_event(self):
        self.assertTrue(self.e._evaluate_code('event.data[\'a\'] == 1', {'event': Event('test', a=1)}))
        self.assertTrue(self.e._evaluate_code('event.name == \'test\'', {'event': Event('test')}))

    def test_execution(self):
        self.e._execute_code('a = 1')
        self.assertEqual(self.e.context['a'], 1)
        self.e._execute_code('x = 2')
        self.assertEqual(self.e.context['x'], 2)

    def test_invalid_condition(self):
        with self.assertRaises(AttributeError):
            self.e._evaluate_code('x.y')

    def test_invalid_action(self):
        with self.assertRaises(AttributeError):
            self.e._execute_code('x = x.y')


from functools import lru_cache


class Event:
    """
    Simple event with a name and (optionally) some data.
    Unless the attribute already exists, each key from ``data`` is exposed as an attribute
    of this class.

    :param name: Name of the event
    :param data: additional data (mapping, dict-like)
    """

    def __init__(self, name: str, **additional_parameters):
        self.name = name
        self.data = additional_parameters

    def __eq__(self, other):
        return isinstance(other, Event) and self.name == other.name

    def __getattr__(self, attr):
        return self.data[attr]

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        if self.data:
            return 'Event({}, {})'.format(self.name, self.data)
        else:
            return 'Event({})'.format(self.name)


class ContractMixin:
    """
    Mixin with a contract: preconditions, postconditions and invariants.
    """

    def __init__(self):
        self._preconditions = []
        self._postconditions = []
        self._invariants = []

    @property
    def preconditions(self):
        """
        A list of preconditions (str).
        """
        return self._preconditions

    @property
    def postconditions(self):
        """
        A list of postconditions (str).
        """
        return self._postconditions

    @property
    def invariants(self):
        """
        A list of invariants (str).
        """
        return self._invariants


class StateMixin:
    """
    State element with a name.

    :param name: name of the state
    """

    def __init__(self, name: str):
        self.name = name

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.name)

    def __eq__(self, other):
        return isinstance(other, StateMixin) and self.name == other.name

    def __hash__(self):
        return hash(self.name)


class ActionStateMixin:
    """
    State that can define actions on entry and on exit.

    :param on_entry: code to execute when state is entered
    :param on_exit: code to execute when state is exited
    """

    def __init__(self, on_entry: str = None, on_exit: str = None):
        self._on_entry = on_entry
        self._on_exit = on_exit

    @property
    def on_entry(self):
        return self._on_entry

    @property
    def on_exit(self):
        return self._on_exit


class TransitionStateMixin:
    """
    A simple state can host transitions
    """

    def __init__(self):
        self._transitions = []

    @property
    def transitions(self):
        return self._transitions


class CompositeStateMixin:
    """
    Composite state can have children states.
    """

    def __init__(self):
        self._children = []

    @property
    def children(self):
        return self._children


class BasicState(ContractMixin, StateMixin, TransitionStateMixin, ActionStateMixin):
    """
    A basic state, with a name, transitions, actions, etc. but no child state.

    :param name: name of this state
    :param on_entry: code to execute when state is entered
    :param on_exit: code to execute when state is exited
    """

    def __init__(self, name: str, on_entry: str = None, on_exit: str = None):
        ContractMixin.__init__(self)
        StateMixin.__init__(self, name)
        TransitionStateMixin.__init__(self)
        ActionStateMixin.__init__(self, on_entry, on_exit)


class CompoundState(ContractMixin, StateMixin, TransitionStateMixin, ActionStateMixin, CompositeStateMixin):
    """
    Compound states must have children states.

    :param name: name of this state
    :param initial: name of the initial state
    :param on_entry: code to execute when state is entered
    :param on_exit: code to execute when state is exited
    """

    def __init__(self, name: str, initial: str = None, on_entry: str = None, on_exit: str = None):
        ContractMixin.__init__(self)
        StateMixin.__init__(self, name)
        TransitionStateMixin.__init__(self)
        ActionStateMixin.__init__(self, on_entry, on_exit)
        CompositeStateMixin.__init__(self)
        self.initial = initial


class OrthogonalState(ContractMixin, StateMixin, TransitionStateMixin, ActionStateMixin, CompositeStateMixin):
    """
    Orthogonal states run their children simultaneously.

    :param name: name of this state
    :param on_entry: code to execute when state is entered
    :param on_exit: code to execute when state is exited
    """

    def __init__(self, name: str, on_entry: str = None, on_exit: str = None):
        ContractMixin.__init__(self)
        StateMixin.__init__(self, name)
        TransitionStateMixin.__init__(self)
        ActionStateMixin.__init__(self, on_entry, on_exit)
        CompositeStateMixin.__init__(self)


class HistoryState(ContractMixin, StateMixin):
    """
    History state can be either 'shallow' (default) or 'deep'.
    A shallow history state resumes the execution of its parent.
    A deep history state resumes the execution of its parent, and of every nested
    active states in its parent.

    :param name: name of this state
    :param initial: name of the initial state
    :param deep: Boolean indicating whether a deep semantic (True) or a shallow semantic (False) should be used
    """

    def __init__(self, name: str, initial: str = None, deep: bool = False):
        ContractMixin.__init__(self)
        StateMixin.__init__(self, name)
        self.name = name
        self.initial = initial
        self.deep = deep


class FinalState(ContractMixin, StateMixin, ActionStateMixin):
    """
    Final state has NO transition and is used to detect state machine termination.

    :param name: name of this state
    :param on_entry: code to execute when state is entered
    :param on_exit: code to execute when state is exited
    """

    def __init__(self, name: str, on_entry: str = None, on_exit: str = None):
        ContractMixin.__init__(self)
        StateMixin.__init__(self, name)
        ActionStateMixin.__init__(self, on_entry, on_exit)


class Transition(ContractMixin):
    """
    A Transition between two states.
    Transition can be eventless or internal (but not both at once).
    A condition (code as string) can be specified as a guard.

    :param from_state: name of the source state
    :param to_state: name of the target state (if transition is not internal)
    :param event: event (if any)
    :param guard: condition as code (if any)
    :param action: action as code (if any)
    """

    def __init__(self, from_state: str, to_state: str = None, event: Event = None, guard: str = None,
                 action: str = None):
        ContractMixin.__init__(self)
        self.from_state = from_state
        self.to_state = to_state
        self.event = event
        self.guard = guard
        self.action = action

    @property
    def internal(self):
        """
        Boolean indicating whether this transition is an internal transition.
        """
        return self.to_state is None

    @property
    def eventless(self):
        """
        Boolean indicating whether this transition is an eventless transition.
        """
        return self.event is None

    def __eq__(self, other):
        return (isinstance(other, Transition) and
                self.from_state == other.from_state and
                self.to_state == other.to_state and
                self.event == other.event)

    def __repr__(self):
        return 'Transition({0}, {1}, {2})'.format(self.from_state, self.to_state, self.event)

    def __str__(self):
        to_state = self.to_state if self.to_state else '[' + self.from_state + ']'
        event = '+' + self.event.name if self.event else ''
        return self.from_state + event + ' -> ' + to_state

    def __hash__(self):
        return hash(self.from_state)


class StateChart(ContractMixin, StateMixin, ActionStateMixin, CompositeStateMixin):
    """
    Python structure for a statechart

    :param name: Name of this statechart
    :param initial: Initial state
    :param on_entry: Code to execute when this statechart is initialized for execution
    """

    def __init__(self, name: str, initial: str, on_entry: str = None):
        ContractMixin.__init__(self)
        StateMixin.__init__(self, name)
        ActionStateMixin.__init__(self, on_entry, None)
        CompositeStateMixin.__init__(self)

        self.initial = initial
        self._states = {}  # name -> State object
        self._parent = {}  # name -> parent.name
        self.transitions = []  # list of Transition objects

    def register_state(self, state: StateMixin, parent):
        """
        Register given state. This method also register the given state
        to its parent.

        :param state: state to add
        :param parent: name of its parent
        """
        # Name should be unused so far
        if state.name in self._states.keys():
            raise ValueError('State name {} is already used!'.format(state.name))
        self._states[state.name] = state
        self._parent[state.name] = parent.name if isinstance(parent, StateMixin) else parent

        # Register on parent state
        parent_state = self._states.get(self._parent[state.name], None)
        if parent_state is not None:
            self._states[self._parent[state.name]].children.append(state.name)
        else:
            # ... or on top-level state (self!)
            self.children.append(state.name)

    def register_transition(self, transition: Transition):
        """
        Register given transition and register it on the source state

        :param transition: transition to add
        """
        self.transitions.append(transition)
        self._states[transition.from_state].transitions.append(transition)

    @property
    def states(self):
        """
        A dictionary that associates a ``StateMixin`` to a state name
        """
        return self._states

    def events(self, state_or_states=None) -> list:
        """
        List of possible event names.
        If *state_or_states* is omitted, returns all possible event names.
        If *state_or_states* is a string, return the events for which a transition exists
        with a *from_state* equals to given string.
        If *state_or_states* is a list of state names, return the events for all those states.

        :param state_or_states: ``None``, a state name or a list of state names.
        :return: A list of event names
        """
        if state_or_states is None:
            states = self._states.keys()
        elif isinstance(state_or_states, str):
            states = [state_or_states]
        else:
            states = state_or_states

        names = set()
        for transition in self.transitions:
            if transition.event and transition.from_state in states:
                names.add(transition.event.name)
        return sorted(names)

    @lru_cache()
    def ancestors_for(self, state: str) -> list:
        """
        Return an ordered list of ancestors for the given state.
        Ancestors are ordered by decreasing depth.

        :param state: name of the state
        :return: state's ancestors
        """
        ancestors = []
        parent = self._parent[state]
        while parent:
            ancestors.append(parent)
            parent = self._parent[parent]
        return ancestors

    @lru_cache()
    def descendants_for(self, state: str) -> list:
        """
        Return an ordered list of descendants for the given state.
        Descendants are ordered by increasing depth.

        :param state: name of the state
        :return: state's descendants
        """
        descendants = []
        states_to_consider = [state]
        while states_to_consider:
            state = states_to_consider.pop(0)
            state = self._states[state]
            if isinstance(state, CompositeStateMixin):
                for child in state.children:
                    states_to_consider.append(child)
                    descendants.append(child)
        return descendants

    @lru_cache()
    def depth_of(self, state: str) -> int:
        """
        Return the depth of given state (0-indexed).

        :param state: name of the state
        :return: state depth
        """
        if state is None:
            return 0
        ancestors = self.ancestors_for(state)
        return len(ancestors) + 1

    @lru_cache()
    def least_common_ancestor(self, s1: str, s2: str) -> str:
        """
        Return the deepest common ancestor for *s1* and *s2*, or ``None`` if
        there is no common ancestor except root (top-level) state.

        :param s1: name of first state
        :param s2: name of second state
        :return: name of deepest common ancestor or ``None``
        """
        s1_anc = self.ancestors_for(s1)
        s2_anc = self.ancestors_for(s2)
        for state in s1_anc:
            if state in s2_anc:
                return state

    def leaf_for(self, states: list) -> list:
        """
        Considering the list of states names in *states*, return a list containing each
        element of *states* such that this element has no descendant in *states*.
        In other words, this method returns the leaves from the given list of states.

        :param states: a list of names
        :return: the names of the leaves in *states*
        """
        leaves = []
        # TODO: Need a more efficient way to compute this set
        for state in states:
            keep = True
            for descendant in self.descendants_for(state):
                if descendant in states:
                    keep = False
                    break
            if keep:
                leaves.append(state)
        return leaves

    def validate(self) -> bool:
        """
        Validate current statechart:

         - C1. Check that transitions refer to existing states
         - C2. Check that history can only be defined as a child of a CompoundState
         - C3. Check that initial state refer to a parent's child
         - C4. Check that orthogonal states have at least one child
         - C5. Check that there is no internal eventless guardless transition

        :return: True if no check fails
        :raise AssertionError: if a check fails
        """
        # C1 & C5
        for transition in self.transitions:
            if (not (transition.from_state in self._states and
                     (not transition.to_state or transition.to_state in self._states))):
                raise AssertionError('Transition {} refers to an unknown state'.format(transition))
            if not transition.event and not transition.guard and not transition.to_state:
                raise AssertionError('Transition {} is an internal, eventless and guardless '
                                     'transition.'.format(transition))

        for name, state in self._states.items():
            if isinstance(state, HistoryState):  # C2 & C3
                if not isinstance(self._states[self._parent[name]], CompoundState):
                    raise AssertionError('History state {} can only be defined in a compound '
                                         '(non-orthogonal) states'.format(state))
                    # Remove because this can be helpful for orthogonal states
                    # if state.initial and not (self._parent[state.initial] == self._parent[name]):
                    #    raise AssertionError('Initial memory of {} should refer to a child of {}'
                    #           .format(state, self._parent[name]))

            if isinstance(state, CompositeStateMixin):  # C4
                if len(state.children) <= 0:
                    raise AssertionError('Composite state {} should have at least one child'.format(state))

            if isinstance(state, CompoundState):  # C3
                if self._parent[name] and state.initial and not (self._parent[state.initial] == name):
                    raise AssertionError('Initial state of {} should refer to one of its children'.format(state))

        return True

    def __repr__(self):
        return 'statechart "{}"'.format(self.name)


class MicroStep:
    """
    Create a micro step. A step consider ``event``, takes ``transition`` and results in a list
    of ``entered_states`` and a list of ``exited_states``.
    Order in the two lists is REALLY important!

    :param event: Event or None in case of eventless transition
    :param transition: a ''Transition`` or None if no processed transition
    :param entered_states: possibly empty list of entered states
    :param exited_states: possibly empty list of exited states
    """

    def __init__(self, event: Event = None, transition: Transition = None,
                 entered_states: list = None, exited_states: list = None):
        self.event = event
        self.transition = transition if transition else []
        self.entered_states = entered_states if entered_states else []
        self.exited_states = exited_states if exited_states else []

    def __repr__(self):
        return 'MicroStep({}, {}, >{}, <{})'.format(self.event, self.transition, self.entered_states, self.exited_states)


class MacroStep:
    """
    A macro step is a list of micro steps.

    :param time: the time at which this step was executed
    :param steps: a list of ``MicroStep`` instances
    """

    def __init__(self, time: int, steps: list):
        self._time = time
        self._steps = steps

    @property
    def steps(self):
        """
        List of micro steps
        """
        return self._steps

    @property
    def time(self):
        """
        Time at which this step was executed.
        """
        return self._time

    @property
    def event(self) -> Event:
        """
        Event (or ``None``) that were consumed.
        """
        for step in self._steps:
            if step.event:
                return step.event
        return None

    @property
    def transitions(self) -> list:
        """
        A (possibly empty) list of transitions that were triggered.
        """
        return [step.transition for step in self._steps if step.transition]

    @property
    def entered_states(self) -> list:
        """
        List of the states names that were entered.
        """
        states = []
        for step in self._steps:
            states += step.entered_states
        return states

    @property
    def exited_states(self) -> list:
        """
        List of the states names that were exited.
        """
        states = []
        for step in self._steps:
            states += step.exited_states
        return states

    def __repr__(self):
        return 'MacroStep@{}({}, {}, >{}, <{})'.format(self.time, self.event, self.transitions,
                                                       self.entered_states, self.exited_states)



# "Export" most used elements
from . import evaluator, interpreter, io, model, stories, testing

__description__ = 'Sismic Interactive State Machine Interpreter and Checker'
__version__ = '0.14.3'
__url__ = 'https://github.com/AlexandreDecan/sismic/'
__author__ = 'Alexandre Decan'
__email__ = 'alexandre.decan@lexpage.net'
__licence__ = 'LGPL3'


import os
import unittest
from sismic import io


class ImportFromYamlTests(unittest.TestCase):
    def test_yaml_tests(self):
        files = ['actions', 'composite', 'deep_history', 'infinite', 'internal', 'nested_parallel',
                 'nondeterministic', 'parallel', 'simple', 'timer']
        for f in files:
            io.import_from_yaml(open(os.path.join('tests', 'yaml', f+'.yaml')))

    def test_examples(self):
        files = ['elevator', 'elevator_contract', 'microwave', 'tester_elevator_7th_floor_never_reached',
                 'tester_elevator_moves_after_10s', 'writer_options']
        for f in files:
            io.import_from_yaml(open(os.path.join('docs', 'examples', f+'.yaml')))


class ExportToDictYAMLTests(unittest.TestCase):
    def test_yaml_tests(self):
        files = ['actions', 'composite', 'deep_history', 'infinite', 'internal', 'nested_parallel',
                 'nondeterministic', 'parallel', 'simple', 'timer']
        for f in files:
            d = io.export_to_yaml(io.import_from_yaml(open(os.path.join('tests', 'yaml', f+'.yaml'))))
            # check
            io.import_from_yaml(d)

    def test_examples(self):
        files = ['elevator', 'elevator_contract', 'microwave', 'tester_elevator_7th_floor_never_reached',
                 'tester_elevator_moves_after_10s', 'writer_options']
        for f in files:
            d = io.export_to_yaml(io.import_from_yaml(open(os.path.join('docs', 'examples', f+'.yaml'))))
            # check
            io.import_from_yaml(d)


import unittest
from sismic import io
from sismic import model

class TraversalTests(unittest.TestCase):
    def setUp(self):
        self.sc = io.import_from_yaml(open('tests/yaml/composite.yaml'))

    def test_ancestors(self):
        self.assertEqual(self.sc.ancestors_for('s2'), [])
        self.assertEqual(self.sc.ancestors_for('s1a'), ['s1'])
        self.assertEqual(self.sc.ancestors_for('s1b1'), ['s1b', 's1'])

    def test_descendants(self):
        self.assertEqual(self.sc.descendants_for('s2'), [])
        s1 = self.sc.descendants_for('s1')
        self.assertTrue('s1a' in s1)
        self.assertTrue('s1b' in s1)
        self.assertTrue('s1b1' in s1)
        self.assertTrue('s1b2' in s1)
        self.assertTrue(len(s1) == 4)

    def test_depth(self):
        self.assertEqual(self.sc.depth_of(None), 0)
        self.assertEqual(self.sc.depth_of('s1'), 1)
        self.assertEqual(self.sc.depth_of('s1a'), 2)
        self.assertEqual(self.sc.depth_of('s1b1'), 3)
        self.assertEqual(self.sc.depth_of('s2'), 1)
        self.assertEqual(self.sc.depth_of('s1b2'), 3)

    def test_lca(self):
        self.assertEqual(self.sc.least_common_ancestor('s1', 's2'), None)
        self.assertEqual(self.sc.least_common_ancestor('s1', 's1a'), None)
        self.assertEqual(self.sc.least_common_ancestor('s1a', 's1b'), 's1')
        self.assertEqual(self.sc.least_common_ancestor('s1a', 's1b1'), 's1')

    def test_leaf(self):
        self.assertEqual(self.sc.leaf_for([]), [])
        self.assertEqual(self.sc.leaf_for(['s1']), ['s1'])
        self.assertEqual(self.sc.leaf_for(['s1', 's2']), ['s1', 's2'])
        self.assertEqual(self.sc.leaf_for(['s1', 's1b1', 's2']), ['s1b1', 's2'])
        self.assertEqual(self.sc.leaf_for(['s1', 's1b', 's1b1']), ['s1b1'])

    def test_events(self):
        self.assertEqual(self.sc.events(), ['click', 'close', 'validate'])
        self.assertEqual(self.sc.events('s1b1'), ['validate'])
        self.assertEqual(self.sc.events(['s1b1', 's1b']), ['click', 'validate'])

    def test_name_collision(self):
        sc = model.StateChart('test', 'initial')
        s1 = model.BasicState('a')
        s2 = model.BasicState('a')
        sc.register_state(s1, parent=None)
        with self.assertRaises(ValueError):
            sc.register_state(s2, parent=None)


import unittest
from sismic import io
from sismic.interpreter import Interpreter
from sismic.stories import Story, Pause, Event
from sismic.testing import teststory_from_trace


class StoryFromTraceTests(unittest.TestCase):
    def test_simple(self):
        interpreter = Interpreter(io.import_from_yaml(open('tests/yaml/simple.yaml')))
        Story([Pause(2), Event('goto s2'), Pause(3)]).tell(interpreter)
        story = teststory_from_trace(interpreter.trace)
        expected = Story([Event('started'),
                          Event('entered', state='s1'),
                          Pause(2),
                          Event('consumed', event=Event('goto s2')),
                          Event('exited', state='s1'),
                          Event('processed', source='s1', target='s2', event=Event('goto s2')),
                          Event('entered', state='s2'),
                          Event('exited', state='s2'),
                          Event('processed', source='s2', target='s3', event=None),
                          Event('entered', state='s3'),
                          Event('stopped')])
        for a, b in zip(story, expected):
            self.assertEqual(a, b)
            if isinstance(a, Event):
                self.assertEqual(a.data.items(), b.data.items())
            else:
                self.assertEqual(a.duration, b.duration)


class ElevatorStoryTests(unittest.TestCase):
    def test_7th_floor_never_reached(self):
        tested = Interpreter(io.import_from_yaml(open('docs/examples/elevator.yaml')))
        story = Story([Event('floorSelected', floor=8)])
        trace = story.tell(tested).trace

        test_story = teststory_from_trace(trace)

        tester = Interpreter(io.import_from_yaml(open('docs/examples/tester_elevator_7th_floor_never_reached.yaml')))
        self.assertTrue(test_story.tell(tester).final)

    def test_7th_floor_never_reached_fails(self):
        tested = Interpreter(io.import_from_yaml(open('docs/examples/elevator.yaml')))
        story = Story([Event('floorSelected', floor=4), Pause(2), Event('floorSelected', floor=7)])
        trace = story.tell(tested).trace

        test_story = teststory_from_trace(trace)

        tester = Interpreter(io.import_from_yaml(open('docs/examples/tester_elevator_7th_floor_never_reached.yaml')))
        self.assertFalse(test_story.tell(tester).final)

    def test_elevator_moves_after_10s(self):
        tested_sc = io.import_from_yaml(open('docs/examples/elevator.yaml'))
        tester_sc = io.import_from_yaml(open('docs/examples/tester_elevator_moves_after_10s.yaml'))

        stories = [
            Story([Event('floorSelected', floor=4)]),
            Story([Event('floorSelected', floor=0)]),
            Story([Event('floorSelected', floor=4), Pause(10)]),
            Story([Event('floorSelected', floor=0), Pause(10)]),
            Story([Event('floorSelected', floor=4), Pause(9)]),
            Story([Event('floorSelected', floor=0), Pause(9)]),
        ]

        for story in stories:
            tested = Interpreter(tested_sc)
            test_story = teststory_from_trace(story.tell(tested).trace)

            tester = Interpreter(tester_sc)
            self.assertTrue(test_story.tell(tester).final)


import unittest
from sismic import io
from sismic.interpreter import Interpreter
from sismic.model import Event, Transition, StateChart, StateMixin
from sismic.testing import PreconditionFailed, PostconditionFailed, InvariantFailed


class ElevatorContractTests(unittest.TestCase):
    def setUp(self):
        self.sc = io.import_from_yaml(open('docs/examples/elevator.yaml'))
        self.interpreter = Interpreter(self.sc)

    def test_no_error(self):
        self.interpreter.send(Event('floorSelected', floor=4))
        self.interpreter.execute()
        self.assertFalse(self.interpreter.final)

    def test_state_precondition(self):
        self.sc.states['movingUp'].preconditions.append('False')
        self.interpreter.send(Event('floorSelected', floor=4))
        with self.assertRaises(PreconditionFailed) as cm:
            self.interpreter.execute()
        self.assertTrue(isinstance(cm.exception.obj, StateMixin))

    def test_state_postcondition(self):
        self.sc.states['movingUp'].postconditions.append('False')
        self.interpreter.send(Event('floorSelected', floor=4))
        with self.assertRaises(PostconditionFailed) as cm:
            self.interpreter.execute()
        self.assertTrue(isinstance(cm.exception.obj, StateMixin))

    def test_state_invariant(self):
        self.sc.states['movingUp'].invariants.append('False')
        self.interpreter.send(Event('floorSelected', floor=4))
        with self.assertRaises(InvariantFailed) as cm:
            self.interpreter.execute()
        self.assertTrue(isinstance(cm.exception.obj, StateMixin))

    def test_transition_precondition(self):
        self.sc.states['floorSelecting'].transitions[0].preconditions.append('False')
        self.interpreter.send(Event('floorSelected', floor=4))
        with self.assertRaises(PreconditionFailed) as cm:
            self.interpreter.execute()
        self.assertTrue(isinstance(cm.exception.obj, Transition))

    def test_transition_postcondition(self):
        self.sc.states['floorSelecting'].transitions[0].postconditions.append('False')
        self.interpreter.send(Event('floorSelected', floor=4))
        with self.assertRaises(PostconditionFailed) as cm:
            self.interpreter.execute()
        self.assertTrue(isinstance(cm.exception.obj, Transition))

    def test_statechart_precondition(self):
        self.sc.preconditions.append('False')
        with self.assertRaises(PreconditionFailed) as cm:
            self.interpreter = Interpreter(self.sc)
        self.assertTrue(isinstance(cm.exception.obj, StateChart))

    def test_statechart_postcondition(self):
        sc = io.import_from_yaml(open('tests/yaml/simple.yaml'))
        sc.postconditions.append('False')
        interpreter = Interpreter(sc)
        interpreter.send(Event('goto s2')).send(Event('goto final'))
        with self.assertRaises(PostconditionFailed) as cm:
            interpreter.execute()
        self.assertTrue(isinstance(cm.exception.obj, StateChart))

    def test_statechart_invariant(self):
        self.sc.invariants.append('False')
        self.interpreter.send(Event('floorSelected', floor=4))
        with self.assertRaises(InvariantFailed) as cm:
            self.interpreter.execute()
        self.assertTrue(isinstance(cm.exception.obj, StateChart))

    def test_do_not_raise(self):
        self.sc.invariants.append('False')
        interpreter = Interpreter(self.sc, ignore_contract=True)
        interpreter.send(Event('floorSelected', floor=4))
        interpreter.execute()